/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.connectivity;

import georouting.ConnectivityContract;
import georouting.Edge;
import georouting.Node;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

/**
 *
 * @author Jordan
 */
public class CDSDisk implements ConnectivityContract
{

  public float connectivityRange() { return radius(); }

  private double _radius;
  
  private Rectangle2D _region;

  public float radius() { return (float) _radius; }
  
  public void setRegion(Rectangle2D region){ _region = region; }

  public CDSDisk(double radius)
  {
    _radius = radius;
    _region = null;
  }

  private ArrayList<Node> setDominator(Node d, List<Node> nodes, List<Edge> edges){
      nodes.remove(d);
      
      Iterator<Node> n_it = nodes.iterator();
      Node n;
      ArrayList<Node> neighbors = new ArrayList<>();
      
      while(n_it.hasNext()){
          n = n_it.next();
          if(connect(d, n, edges))
          {            
            n_it.remove();
            neighbors.add(n);
          }          
      }
      return neighbors;
  }
  
  private boolean connect(Node a, Node b, List<Edge> edges){
      if(a.distance(b) <= radius() )
      {
          try
          {
              edges.add(new Edge(a,b));
          }
          catch(Exception e)
          { 
              System.err.println(
                  "ERROR: you just tried to justify the existence of an Imaginary Node. STOP THAT!"); 
          }
          return true;
      }
      return false;
  }
  
  public List<Edge> constructionConnectivity(List<Node> nodes)
  {
    Random rng = new Random();
    ArrayList<Node> available = new ArrayList<>();
    for(Node n : nodes)
        if(_region.contains(n.point()))
            available.add(n);

    ArrayList<Edge> rtn = new ArrayList<Edge>();
    ArrayList<Node> dominators = new ArrayList<>();
    HashMap<Node, ArrayList<Node> > clusters = new HashMap<>();
    
    while(!available.isEmpty()){
        Node dominator = available.get(rng.nextInt(available.size()));
        dominators.add(dominator);
        clusters.put(dominator, setDominator(dominator, available, rtn));
    }
        
    for(int i = 0; i < dominators.size(); i++){
        for(int j = i + 1; j < dominators.size(); j++){
            boolean connected = false;
            // Check for 1-hop connection
            // Check my neighbors against them
            for(Node n_i : clusters.get(dominators.get(i))){
                if((connected = connect(n_i, dominators.get(j), rtn)))
                    break;
            }
            if(connected) continue;

            // Check me against their neighbors
            for(Node n_j : clusters.get(dominators.get(j))){
                if((connected = connect(n_j, dominators.get(i), rtn)))
                    break;
            }
            if(connected) continue;
            
            
            // If fail, check for 2-hop connection
            for(Node n_i : clusters.get(dominators.get(i))){
                for(Node n_j : clusters.get(dominators.get(j))){
                    if(connect(n_i, n_j, rtn))
                        break;
                }
            }
                    
        }        
    }
    
    for(Node a : nodes)
    {
      for(Node b : nodes)
      {
        if( a.id() < b.id() && a.distance(b) <= radius() )
        {
          try
          {
            rtn.add(new Edge(a,b));
          }
          catch(Exception e)
          { 
            System.err.println(
              "ERROR: you just tried to justify the existence of an Imaginary Node. STOP THAT!"); 
          }
        }
      }
    }
    
    return rtn;
  }

  public int densityToNumberOfNodes(double density, double width, double height)
  {
    return (int) (
        (density*width*height)/
        (Math.PI*_radius*_radius));
  }

}
