/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.simulation;

import georouting.ConnectivityContract;
import georouting.graph.CFRGraph;
import georouting.connectivity.StaticUnitDiskConnectivity;
import georouting.planarization.*;
import georouting.traversalAlgorithms.*;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.TraversalAlgorithm;
import georouting.Visualizer;
import georouting.connectivity.CDSDisk;
import georouting.graph.StaticGraph;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

public class StaticSimCFRG {
    // Parameters
    private static final int computations_ = 1;
    private static final double unitDistance_ = 50.0;
    private static final ConnectivityContract cc_ = new StaticUnitDiskConnectivity(unitDistance_);
    private static final PlanarizationAlgorithm algo_ = new GabrialGraph();
    private static final PostOffice po_ = PostOffice.instance();
    
    // Utilities
    private static final Random rand_ = new Random();
    private static final DataSet overhead_ = new DataSet();
    private static final DataSet delivery_ = new DataSet();
    private static final DataSet latency_ = new DataSet();
    
    public static void main(String[] args){
        TraversalAlgorithm  message;
        TraversalAlgorithm  messenger;
        StaticGraph         graph;
        Node                source;
        Node                dest;
        PrintWriter         out = null;
        PrintWriter         log = null;
        
        Visualizer malfunction;
        
        try {            
            FileWriter fw = new FileWriter("BM-results.txt", false);
            FileWriter fw2 = new FileWriter("BM-progress.txt", false);
            out = new PrintWriter(fw);
            log = new PrintWriter(fw2);
        } catch (IOException ex) {
            Logger.getLogger(StaticSim.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        // Print Header
        out.println("Area\tOverhead\tLatency\t\tDelivery");
        out.println("\tAverage\tSt.Dev\tAverage\tSt.Dev\tAverage\tSt.Dev");
        for(int i = 0; i < 40; i++)
            out.print("-");
        out.println();
        
        for(double density = 7; density <= 7; density++){
            for(double s = 1; s <= 1; s++){
                for(double area = 10; area <= 10; area += 1){
                    double bound = (area - s) * unitDistance_;
                    log.println("density " + Double.toString(density) + "/20");
                    System.out.println("density " + Double.toString(density) + "/20");
//                  log.println("area " + Double.toString(s) + "/10");
//                  System.out.println("area " + Double.toString(s) + "/10");
//                  log.println("size " + Double.toString(area) + "/40");
//                  System.out.println("size " + Double.toString(area) + "/40");
                    for(int comp = 0; comp < computations_; comp++){                
                        do{                                                         
                            BM.setCastingArea(rand_.nextDouble() * bound, rand_.nextDouble() * bound,
                                                s * unitDistance_);
                            //((CDSDisk)cc_).setRegion(BM.castArea());
                            graph = new CFRGraph(density, area * unitDistance_, area * unitDistance_, cc_);
                            source = graph.nodes().get(0);
                            //dest = graph.nodes().get(rand_.nextInt(graph.nodes().size() - 1) + 1);

                            graph.planarize(algo_);
                        
                            dest = null;
                            for(Node n : graph.nodes()){
                                if(BM.inArea(n) && n != graph.nodes().get(0) && graph.isConnected(source, n)){
                                    dest = n;
                                    break;
                                }
                            }
                        }while(source.neighborhood().size() == 0 || dest == null); //!graph.isConnected(source, dest));
//                        System.out.println("Area before greed:\t" + BM.castArea().toString());
//                    System.out.println("Targets:");
                        for(Node n : graph.nodes()){
                            if(BM.inArea(n) && graph.isConnected(source, n)){
                                BM.addTarget(n);
//                                System.out.println(n.id() + " at " + n.point().x + ", " + n.point().y);
                            }
                        }
                    
                        int castSize = BM.castSize();
                    
                        po_.reset();
                        //CFR.winningPath_ = null;

                        messenger = new GFG(graph, source, BM.areaCenter(), null);                         
                        
//                        Visualizer vmessenger = new Visualizer(messenger);
//                        vmessenger.begin();
                        
                        while(!BM.inArea(messenger.current()) && !messenger.done()){
                            messenger.advance();
                            //System.out.println("Curr: " + messenger.current().id());
                        }

                        BM.increaseStep(messenger.hops());
//                        System.out.println("GFG: " + messenger.hops());
                        
//                      System.out.println("Delivered to: " + messenger.current().id());
//                      System.out.println(messenger.hops());                                              
//                        System.out.println("Area after greed:\t" + BM.castArea().toString());
                        
                        dest = BM.areaCenter();
//                        message = po_.generate(new BM(graph, source, dest, algo_));
                        message = po_.generate(new BM(graph, messenger.current(), dest, algo_));
                        ((BM)message).initiate();
                        
                        malfunction = new Visualizer(message);
                        malfunction.begin();
                        
//                      System.out.println("Initial:\n" + po_);
                        if(comp % 50 == 0){
                            log.println("comp " + Integer.toString(comp + 1) + "/" + Integer.toString(computations_));
                            System.out.println("comp " + Integer.toString(comp + 1) + "/" + Integer.toString(computations_));
                            log.flush();
                        }
                        
                        // Run execution
                        while(!po_.isEmpty()){
                            BM.increaseStep(1);
                            po_.advance();
//                          System.out.println(po_);
//                          log.println(po_);
                        }
//                      System.out.println("Missed:");
//                      System.out.println("");
//                      System.out.println(BM.delivered() ? "Success" : "Failure");
                        
//                        malfunction.end();
//                        malfunction.dispose();
                        System.out.println("packets: " + po_.packets());
                        System.out.println("latency: " + BM.latency());
                        
                        // Record Results
//                        overhead_.add(loss, (double)(BM.visited()+BM.broadcasts()));

                        if(BM.latency() != -1){
                            overhead_.add(density, (double)(po_.packets()));// + messenger.hops()));
                            delivery_.add(density, (double)(castSize - BM.castSize()) / castSize);
                            latency_.add(density, BM.latency());
//                            malfunction.end();
//                            malfunction.dispose();
                        }
//                        else{
//                            System.out.println("Delivery probs again");
//                            return;
//                        }
//                        System.out.println(messenger.hops() + BM.visited()+BM.broadcasts());
//                        System.out.println(messenger.hops());
//                    System.out.println("packets: " + po_.packets());

//                        System.out.println("Result:\n" + (double)(castSize - BM.castSize()) / castSize);
                    
                    
                    }
                    // Write results to file
                    out.print(density);
                    out.print("\t");
                    out.print(overhead_.average(density));
                    out.print("\t");
                    out.print(overhead_.stDev(density));
                    out.print("\t");
                    out.print(latency_.average(density));
                    out.print("\t");
                    out.print(latency_.stDev(density));
                    out.print("\t");
                    out.print(delivery_.average(density));
                    out.print("\t");
                    out.print(delivery_.stDev(density));
                    out.println();
                    out.flush();                
//              System.out.println("Progress: " + Double.toString(density) + "/20");
                }                
            }
        }                       
        log.println("Finished!");
        System.out.println("Finished!");
        
        out.close();        
        log.close();
    }
}
