/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class BM extends PlanarTraversalAlgorithm{
    protected static Rectangle2D castArea_;
    protected static ArrayList<Node> castTargets_ = new ArrayList<>();
    protected static ArrayList<Node> visited_ = new ArrayList<>();
    protected static int broadcasts_ = 0;
    protected static double loss_ = 0;
    protected static double latency_;
    protected static double steps_;
    protected boolean discarded_ = false;
    protected static ImaginaryNode p_;
    protected Node previous_;    
    protected Node start_;
    protected Node prevStart_;   
    
    public BM(Graph g, Node s, Node f, PlanarizationAlgorithm pa){
        super(g, s, f, pa);
        start_ = s;        
    }
    
    public static void setCastingArea(double x, double y, double sideLength){        
        castArea_ = new Rectangle2D.Double(x, y, sideLength, sideLength);
        castTargets_.clear();
        visited_.clear();
        broadcasts_ = 0;
        latency_ = -1;
        steps_ = 0;
    }
    
    @Override
    public String name(){
        return "BM";
    }
    
    protected void visit(Node n){
        if(castTargets_.remove(n)){
            if(castTargets_.isEmpty()){
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(_container, _container.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while(!optimal.done()) optimal.advance();
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }
                
            }
        }
        if(!visited_.contains(n))
            visited_.add(n);
    }
    
    @Override
    public boolean done(){
        return discarded_;
    }
    
    @Override
    protected void advanceAlgorithm()
    {
        if(!done()){
            visit(_current);
            
            if(previous_ == null){
                discarded_ = true;
                //System.out.println("Previous was found to be NULL during traversal:");
                //System.out.println(toString());
            }
            
            Node target = bestNeighbor();            
            PostOffice.instance().addPacket(); // Added as we jump to target to check (current, target)
            
            if(isEntry(_current, target) || isEntry(target, _current)){
                //System.out.println("-Changed faces-");
                // Change faces
                previous_ = target;
                target = bestNeighbor();
            }
            else{
                //System.out.println("T: " + target.id() + ", " + toString());
                moveTo(target);         
            }                                    
            
            if(_current == start_ && previous_ == prevStart_)
                discarded_ = true;
        }
    }
    
    public void initiate()
    {
        // Get start edge
        previous_ = areaCenter();

        prevStart_ = bestCWNeighbor();
        previous_ = prevStart_;
        
        float p_x = (float)((_current.x() + previous_.x()) * 0.45 + 0.1 * areaCenter().x());
        float p_y = (float)((_current.y() + previous_.y()) * 0.45 + 0.1 * areaCenter().y());
        
        p_ = new ImaginaryNode(p_x, p_y);
        
        visit(_current);
        
        //System.out.println("E-Start: [" + prevStart_.id() + ", " + start_.id() + "]");
    }
    
    protected boolean isEntry(Node u, Node v){
        BMScout scout = new BMScout(u, v);
        do{
            scout.advance();           
        }while(!scout.done());   
        
        steps_ += scout.steps();        
        
        return scout.entry().u() == u && scout.entry().v() == v && scout.regional(); // && scout.comfirm(u, v);
    }
    
    protected class BMScout {
        Node start_;
        Node startPrev_;
        Node curr_;
        Node prev_;
        Key entry_;
        boolean regional_;
        int hops_;
        
        BMScout(Node u, Node v){
            start_ = curr_ = v;
            startPrev_ = prev_ = u;
            entry_ = new Key(u, v);
            regional_ = false;
            hops_ = 0;
        }
        
        public boolean regional(){
            return regional_;
        }
        
        public Key entry(){
            return entry_;
        }
        
        public void advance(){
            if(inArea(curr_))
                regional_ = true;
            
            Node target = next();
            move(target);
            Key key = new Key(prev_, curr_);
            
            if(key.compare(entry_))
                entry_ = key;
        }
       
        public boolean comfirm(Node u, Node v){
            boolean legitimate = true;
            
            do{
                if(curr_ == u && prev_ == v){
                    //System.out.println("Disproved Entry [" + u.id() + ", " + v.id() + "]");
                    legitimate = false;
                    break;
                }
                
                advance();                
            }while(!done());
            
            return legitimate;
        }
        
        public boolean done(){
            return curr_ == start_ && prev_ == startPrev_;
        }
        
        protected void move(Node n){
            hops_++;
            prev_ = curr_;
            curr_ = n;
            PostOffice.instance().addPacket();
        }
    
        public int steps(){
            return hops_;
        }
        
        protected Node next()
        {
            List<Node> neighborhood = curr_.neighborhood();
            Node rtn = prev_;
            double minAngle = Double.MAX_VALUE;
            for(Node n : neighborhood)
            {
                float angle = getCCWAngle(prev_, curr_, n);
                if(angle > 0 && angle < minAngle)
                {
                  minAngle = angle;
                  rtn = n;
                }
            }
            return rtn;
        }    
    }
    
    
    protected void moveTo(Node n){
        _hops++;
        previous_ = _current;
        _current = n;
        PostOffice.instance().addPacket();
    }
    
    protected Node bestNeighbor()
    {
        List<Node> neighborhood = _current.neighborhood();
        Node rtn = previous_;
        double minAngle = Double.MAX_VALUE;
        for(Node n : neighborhood)
        {
            float angle = getCCWAngle(previous_, _current, n);
            if(angle > 0 && angle < minAngle)
            {
              minAngle = angle;
              rtn = n;
            }
        }
        return rtn;
    }
    
    protected Node bestCWNeighbor()
    {
        List<Node> neighborhood = _current.neighborhood();
        Node rtn = previous_;
        double minAngle = Double.MAX_VALUE;
        for(Node n : neighborhood)
        {
            float angle = getCWAngle(previous_, _current, n);
            if(angle > 0 && angle < minAngle)
            {
              minAngle = angle;
              rtn = n;
            }
        }
        return rtn;
    }
    
    protected float getDirection(ImaginaryNode pt)
    {    
        float c_x = p_.x() + 50;
        float c_y = p_.y();
                
        double angle = Math.atan2( c_y - p_.y() , c_x - p_.x() )
                 - Math.atan2( pt.y() - p_.y() , pt.x() - p_.x() );
        
        if(angle < 0.0)
            angle = 2 * Math.PI + angle;
        
        return (float) angle;
    }
    
    protected float getCCWAngle(Node a, Node b, Node c)
    {
        double angle = Math.atan2( c.y() - b.y() , c.x() - b.x() )
                     - Math.atan2( a.y() - b.y() , a.x() - b.x() );
        if(angle < 0.0)
            angle = 2 * Math.PI + angle;
        
        

        return (float)angle;
    }
    
    protected float getCWAngle(Node a, Node b, Node c)
    {
        double angle = 2 * Math.PI - getCCWAngle(a, b, c);

        return (float)angle;
    }
    
    protected float getAngle(Node a, Node b, Node c)
    {
        double angle = getCCWAngle(a, b, c);
        
        if(angle > Math.PI)
            angle = 2 * Math.PI - angle;

        return (float)angle;
    }
    
    protected ImaginaryNode getPt(Node u, Node v){
        // equation of line through points u and v
        double m = (v.y() - u.y())/(v.x() - u.x());
        double b = u.y() - m * u.x();
        
        double m_p = 1.0 / m * -1;
        double b_p = p_.y() - m_p * p_.x();
        
        double x_pt = (b_p - b) / (m - m_p);
        double y_pt = x_pt * m + b;
        
        // Confirm that pt lies on the edge
        if(!(Math.min(u.x(), v.x()) <= x_pt && x_pt <= Math.max(u.x(), v.x()))){
            if(u.distance(p_) < v.distance(p_)){
                x_pt = u.x();
                y_pt = u.y();
            }
            else{
                x_pt = v.x();
                y_pt = v.y();
            }
        }
                    
        return new ImaginaryNode(x_pt, y_pt);
    }
    
    class Key{
        float first_;
        float second_;
        float third_;
        float fourth_;
        Node u_;
        Node v_;
        
        public Key(Node u, Node v){
            ImaginaryNode pt = getPt(u, v);
            first_ = pt.distance(p_);
            second_ = getDirection(pt);
            third_ = getAngle(p_, u, v);
            fourth_ = getCCWAngle(p_, u, v);
            u_ = u;
            v_ = v;
        }
        
        public boolean compare(Key key){
            // Returns lexicographic <=
            if(first_ != key.first_)
                return first_ <= key.first_;
            if(second_ != key.second_)
                return second_ <= key.second_;
            if(third_ != key.third_)
                return third_ <= key.third_;
            if(fourth_ != key.fourth_)
                return fourth_ <= key.fourth_;
            
            if(u_ != key.u_ || v_ != key.v_){
                //System.out.println("Totally equivalent keys:");
                //System.out.println("\t" + toString());
                //System.out.println("\t" + key);  
            }
                      
            return true;
        }
        
        public Node u(){
            return u_;
        }
        
        public Node v(){
            return v_;
        }
        
        @Override
        public String toString(){
            String str = "[" + Integer.toString(u_.id()) + ", " + Integer.toString(v_.id()) + "] => {"
                    + Float.toString(first_) + ", " + Float.toString(second_) + ", "
                    + Float.toString(third_) + ", " + Float.toString(fourth_) + "}";
            
            return str;
        }       
    }
                   
    public static void increaseStep(double amount){        
        steps_ += amount;
//        //System.out.println("steps: " + steps_);
    }
    
    public static double latency(){
        return latency_;
    }
    public static void setLoss(double loss){
        loss_ = loss;
    }
    
    public static boolean inArea(Node n){
        return castArea_.contains(n.point());
    }      
     
    @Override
    public void visPaint(Graphics g)
    {
        super.visPaint(g);
                
        p_.paint(g, Color.magenta);
        
        g.setColor(Color.cyan);
        g.drawRect((int)castArea_.getX(), (int)castArea_.getY(),
                (int)castArea_.getWidth(), (int)castArea_.getHeight());
        
        canvas().repaint();
    }
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);         
    }
    
    public static void addTarget(Node n){
        castTargets_.add(n);
    }
    
    public static boolean delivered(){
        
        if(!castTargets_.isEmpty()){
            for(Node n : castTargets_){
                //System.out.println(Integer.toString(n.id()));                
            }            
        }
        return castTargets_.isEmpty();
    }
    
    public static Rectangle2D castArea(){
        return castArea_;
    }
    
    public static int castSize(){
//        for(Node n : castTargets_){
//                System.out.println(Integer.toString(n.id()));                
//        }
        return castTargets_.size();
    }
        
    public static int visited(){
        return visited_.size();
    }
    
    public static int broadcasts(){
        return broadcasts_;
    }
    
    @Override
    public String toString(){
        String str = "S: " + Integer.toString(_source.id())
                + ", P: " + Integer.toString(previous_ == null ? -1 : previous_.id())
                + ", C: " + Integer.toString(_current.id()) + ", D: {";
        for(Node dest : castTargets_){
            str += Integer.toString(dest.id()) + " ";
        }
        str += "}";
        
        return str;
    }
    
    public static ImaginaryNode areaCenter(){
        return new ImaginaryNode(castArea_.getCenterX(), castArea_.getCenterY());
    }
}

