/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class CFR extends FACE{
    protected Node                  target_;
    static int                      seqNo_;
    int                             id_;
    private boolean                 discarded_;
    protected ArrayList<Node>       path_;
    public static ArrayList<Node>   winningPath_;
    
    public CFR(Graph g, Node s, Node f, PlanarizationAlgorithm pa){
        super(g, s, f, pa);
        id_ = ++seqNo_;
        discarded_ = false;
        path_ = new ArrayList<>();
        path_.add(s);
    }
    
    public CFR(Graph g, Node s, Node f, Node c, Node t, int id, ArrayList<Node> path, PlanarizationAlgorithm pa){
        super(g, s, f, pa);
        _current = c;
        target_ = t;
        id_ = id;
        discarded_ = false;
        path_ = (ArrayList<Node>)path.clone();
    }
    
  public int id(){
      return id_;
  }
    
  @Override
  protected void advanceAlgorithm()
  {
    if(!_current.neighborhood().isEmpty() && !done()){
        if(!((CFRNode)target_).hasMate(this)){
            moveTo(target_);
            target_ = bestNeighbor();
            
            Edge tEdge = null;
            try{ tEdge = new Edge(_current, target_); }
            catch(Exception e) { System.err.println("Critical Error"); System.exit(1); }
            
            if(justTraversed().intersects(xd()) || tEdge.intersects(xd())
                    || xd().intersects(target_) || xd().intersects(_current))
                split(_previous);
        }
        else{
            PostOffice.instance().addPacket();        
            discard();
        }
    }
  }
  
  public void initiate()
  {
    Node left, right;
      
    if(_current.neighborhood().size() > 1){
        TreeMap<Double, Node> sorted = new TreeMap<>();
        double zero = getAngle(_destination);
      
        for(Node n : _current.neighborhood()){
            double angle = getAngle(n) - zero;
            if(angle < 0) angle += 2 * Math.PI;   
            sorted.put(angle, n);
        }
              
        if(sorted.firstKey() > 0.0){
            left = sorted.lastEntry().getValue();
            right = sorted.firstEntry().getValue();
        }
        else if(_current.neighborhood().size() > 2){
            // On S-D line and multiple secondary winners to choose from
            Node sd = sorted.firstEntry().getValue();
            sorted.remove(sorted.firstKey());

            if(2 * Math.PI - sorted.lastKey() < sorted.firstKey()){
                left = sorted.lastEntry().getValue();
                right = sd;
            }
            else{
                left = sd;
                right = sorted.firstEntry().getValue();
            }
        }
        else{
            // On S-D Line and single secondary
            left  = (sorted.lastKey() > Math.PI ? sorted.lastEntry().getValue() : sorted.firstEntry().getValue());
            right = (sorted.lastKey() > Math.PI ? sorted.firstEntry().getValue() : sorted.lastEntry().getValue());
        }
    }
    else{
        left = _current.neighborhood().get(0);
        right = left;
    }        
        
    target_ = left;
    _dir = Direction.CLOCKWISE;
    CFR mimic = new CFR(_container, _source, _destination, _current, right, id_, path_, _pAlg);
    mimic._dir = Direction.COUNTERCLOCKWISE;
    
    PostOffice.instance().generate(mimic);
    ((CFRNode)_current).receive(this);
    ((CFRNode)_current).receive(mimic);
  }
  
  @Override
  public String toString(){
      return "{id: " + Integer.toString(id_) + ", c: " + Integer.toString(_current.id())
        + ", t: " + Integer.toString(target_.id()) + ", d: " + Integer.toString(_destination.id())
            + ", " + (_dir == Direction.CLOCKWISE ? "R" : "L") + (discarded_ ? "} -D" : "}");
  }
  
  public boolean isMate(CFR message){
      return ((target_ == message._current)
              && (_current == message.target_)
              && (!message.done() && !done())
              && (message._dir != _dir)
              && (_destination == message._destination)
              );
  }
  
  public void discard(){
      CFRNode cur = (CFRNode)_current;
      cur.send(this);
      discarded_ = true;
      //System.out.println("Removed: " + this);
  }
  
  @Override
  public boolean done(){
      return discarded_;
  }
  
  @Override
  protected void moveTo(Node n){
      CFRNode current = (CFRNode)_current;
      CFRNode target = (CFRNode)n;
      
      current.send(this);
      super.moveTo(n);
      target.receive(this);
      PostOffice.instance().addPacket();
      path_.add(0, n);
      
      if(successful() && winningPath_ == null){
          PostOffice.instance().deliver(this);
          winningPath_ = (ArrayList<Node>)path_.clone();
      }
  }
  
  protected double getAngle(Node r){
      return Math.atan2((r.y() - _current.y()), (r.x() - _current.x()));
  }
  
  protected TreeMap<Double, Node> getASL(Node zeroNeighbor){
      TreeMap<Double, Node> sorted = new TreeMap<>();
      double zero = getAngle(zeroNeighbor);
      
      for(Node n : _current.neighborhood()){
          double angle = getAngle(n) - zero;
          if(angle < 0) angle += 2 * Math.PI;
          
          if(_dir == Direction.CLOCKWISE)
              sorted.put(-1 * angle, n);
          else
              sorted.put((n == zeroNeighbor ? 10.0 : angle), n);
      }
      
      return sorted;
  }
  
  public void split(Node origin){      
      TreeMap<Double, Node> asl = getASL(origin);
      
      Iterator<Node> first  = asl.values().iterator();
      Iterator<Node> second = asl.values().iterator();
      second.next();
        
      while(second.hasNext()){
          Node left  = (_dir == Direction.CLOCKWISE ? second.next() : first.next());
          Node right = (_dir == Direction.CLOCKWISE ? first.next() : second.next());
          
          Edge lEdge = null;
          Edge rEdge = null;
          try{ lEdge = new Edge(_current, left); }
          catch(Exception e) { System.err.println("Critical Error"); System.exit(1); }
          try{ rEdge = new Edge(_current, right); }
          catch(Exception e) { System.err.println("Critical Error"); System.exit(1); }
          
          if(lEdge.intersects(xd()) || rEdge.intersects(xd())){                           
              // Send Concurrent Messages
              CFR lSplit = new CFR(_container, _source, _destination, _current, left, id_, path_, _pAlg);
              CFR rSplit = new CFR(_container, _source, _destination, _current, right, id_, path_, _pAlg);
              lSplit._dir = Direction.CLOCKWISE;
              rSplit._dir = Direction.COUNTERCLOCKWISE;
              PostOffice.instance().add(lSplit);
              PostOffice.instance().add(rSplit);
              ((CFRNode)_current).receive(lSplit);
              ((CFRNode)_current).receive(rSplit);
          }
      }
  }
  
    @Override
    public void flipDirection(){
        if(_dir == Direction.CLOCKWISE)
            _dir = Direction.COUNTERCLOCKWISE;
        else
           _dir = Direction.CLOCKWISE; 
    }
    
    @Override
    public String name(){
        return "CFR";
    }
}


