/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class CFRG extends CFR{
    protected static Rectangle2D castArea_;
    protected static ArrayList<Node> castTargets_ = new ArrayList<>();
    protected static ArrayList<Node> visited_ = new ArrayList<>();
    protected static int broadcasts_ = 0;
    protected static double loss_ = 0;
    protected static double latency_;
    protected static double steps_;
    
    public CFRG(Graph g, Node s, Node f, PlanarizationAlgorithm pa){
        super(g, s, f, pa);
        seqNo_ = 0;
        id_ = 0;        
    }
    
    public CFRG(Graph g, Node s, Node f, Node c, Node t, int id, ArrayList<Node> path, PlanarizationAlgorithm pa){
        super(g, s, f, c, t, id, path, pa);
    }
    
    public static void setCastingArea(double x, double y, double sideLength){        
        castArea_ = new Rectangle2D.Double(x, y, sideLength, sideLength);
        castTargets_.clear();
        visited_.clear();
        broadcasts_ = 0;
        latency_ = -1;
        steps_ = 0;
    }
    
    @Override
    public String name(){
        return "CFRG";
    }
    
    protected void visit(Node n){
        if(castTargets_.remove(n)){
            if(castTargets_.isEmpty()){
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(_container, _container.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while(!optimal.done()) optimal.advance();
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }
                
            }
        }
        if(!visited_.contains(n))
            visited_.add(n);
    }
    
    @Override
    protected void advanceAlgorithm()
    {
        if(!done()){
            if(!((CFRNode)target_).hasMate(this)){
                /*
                if(rand_.nextDouble() < loss_){
                    System.out.println(this + " discarded");
                    discard();

                }
                */
                {                    
                    moveTo(target_);
                    visit(_current);
                    target_ = bestNeighbor();                    
                    
                    if(juncture(_previous) || juncture(target_))
                        split(_previous);            
                }
            }
            else{
                PostOffice.instance().addPacket();
                broadcasts_++;
                discard();
            }
        }
    }
    
    @Override
    public void initiate()
    {
        Node left, right;
      
        if(_current.neighborhood().size() > 1){
            TreeMap<Double, Node> sorted = new TreeMap<>();
            double zero = getAngle(_destination);
      
            for(Node n : _current.neighborhood()){
                double angle = getAngle(n) - zero;
                if(angle < 0) angle += 2 * Math.PI;   
                sorted.put(angle, n);
            }
              
            if(sorted.firstKey() > 0.0){
                left = sorted.lastEntry().getValue();
                right = sorted.firstEntry().getValue();
            }
            else if(_current.neighborhood().size() > 2){
                // On S-D line and multiple secondary winners to choose from
                Node sd = sorted.firstEntry().getValue();
                sorted.remove(sorted.firstKey());

                if(2 * Math.PI - sorted.lastKey() < sorted.firstKey()){
                    left = sorted.lastEntry().getValue();
                    right = sd;
                }
                else{
                    left = sd;
                    right = sorted.firstEntry().getValue();
                }
            }
            else{
                // On S-D Line and single secondary
                left  = (sorted.lastKey() > Math.PI ? sorted.lastEntry().getValue() : sorted.firstEntry().getValue());
                right = (sorted.lastKey() > Math.PI ? sorted.firstEntry().getValue() : sorted.lastEntry().getValue());
            }
        }
        else{
            left = _current.neighborhood().get(0);
            right = left;
        }        
        
        target_ = left;
        _dir = Direction.CLOCKWISE;
        CFRG mimic = new CFRG(_container, _source, _destination, _current, right, id_, path_, _pAlg);
        mimic._dir = Direction.COUNTERCLOCKWISE;
    
        PostOffice.instance().generate(mimic);
        ((CFRNode)_current).receive(this);
        ((CFRNode)_current).receive(mimic);
        
        split(right);
        PostOffice.instance().merge();
        
        visit(_current);
    }
    
    @Override
    public void split(Node origin){
        if(_current.neighborhood().size() > 1){
            TreeMap<Double, Node> sorted = new TreeMap<>();
            double zero = getAngle(origin);
                
            for(Node n : _current.neighborhood()){
                if(!sorted.containsValue(n)){
                    Double angle = getAngle(n) - zero;
                    if(angle < 0.0) angle += 2 * Math.PI;
                    sorted.put(angle, n);
                }
            }
            if(_dir == Direction.COUNTERCLOCKWISE)
                sorted.put(10.0, sorted.remove(0.0));
        
            Iterator<Node> first  = sorted.values().iterator();
            Iterator<Node> second = sorted.values().iterator();
            second.next();
        
            CFRNode current = (CFRNode)_current;

            while(second.hasNext()){
                Node left  = first.next();
                Node right = second.next();
                
                if(juncture(left) || juncture(right)){
                    CFRG fwdLeft  = new CFRG(_container, _source, _destination, _current, left, ++seqNo_, path_, _pAlg);
                    CFRG fwdRight = new CFRG(_container, _source, _destination, _current, right, ++seqNo_, path_, _pAlg);
                    fwdRight.flipDirection();
                
                    current.receive(fwdLeft);
                    current.receive(fwdRight);
                    PostOffice.instance().add(fwdLeft);
                    PostOffice.instance().add(fwdRight);                   
                }
            }
        }
    }

    public static void increaseStep(double amount){        
        steps_ += amount;
//        System.out.println("steps: " + steps_);
    }
    
    public static double latency(){
        return latency_;
    }
    public static void setLoss(double loss){
        loss_ = loss;
    }
    
    public static boolean inArea(Node n){
        return castArea_.contains(n.point());
    }
    
    public boolean juncture(Node target){
        Edge tEdge = null;       
        try{ tEdge = new Edge(_current, target); }
        catch(Exception e) { System.err.println("Critical Error"); System.exit(1); }
        
        return(tEdge.intersects(xd()) || inArea(target) || inArea(_current));
    }
     
    @Override
    public void visPaint(Graphics g)
    {
        super.visPaint(g);
        
        g.setColor(Color.cyan);
        g.drawRect((int)castArea_.getX(), (int)castArea_.getY(),
                (int)castArea_.getWidth(), (int)castArea_.getHeight());
                    
        canvas().repaint();
    }
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);         
    }
    
    public static void addTarget(Node n){
        castTargets_.add(n);
    }
    
    public static boolean delivered(){
        
        if(!castTargets_.isEmpty()){
            for(Node n : castTargets_){
                System.out.println(Integer.toString(n.id()));                
            }            
        }
        return castTargets_.isEmpty();
    }
    
    public static Rectangle2D castArea(){
        return castArea_;
    }
    
    public static int castSize(){
//        for(Node n : castTargets_){
//                System.out.println(Integer.toString(n.id()));                
//        }
        return castTargets_.size();
    }
        
    public static int visited(){
        return visited_.size();
    }
    
    public static int broadcasts(){
        return broadcasts_;
    }
    
    public static ImaginaryNode areaCenter(){
        return new ImaginaryNode(castArea_.getCenterX(), castArea_.getCenterY());
    }
}

